//
//  FSEventsService.swift
//  fsevents-example
//
//  Created by Michael Ovchinnikov on 06/02/2019.
//  Copyright © 2019 Michael Ovchinnikov. All rights reserved.
//

import Foundation

class FSEventsService {
    var callback: ((FSEvent) -> Void)?
    let pathsToWatch: [String]
    
    private var stream: FSEventStreamRef!
    
    private let eventCallback: FSEventStreamCallback = {
        (_: ConstFSEventStreamRef,
         contextInfo: UnsafeMutableRawPointer?,
         numEvents: Int,
         eventPaths: UnsafeMutableRawPointer,
         eventFlags: UnsafePointer<FSEventStreamEventFlags>,
         eventIds: UnsafePointer<FSEventStreamEventId>) in
        
        guard let contextInfo = contextInfo else { return }
        let mySelf = Unmanaged<FSEventsService>.fromOpaque(contextInfo).takeUnretainedValue()
        
        guard
            let callback = mySelf.callback,
            let paths = unsafeBitCast(eventPaths, to: NSArray.self) as? [String]
        else {
            return
        }
        
        var events: [FSRawEvent] = []
        for index in 0 ..< numEvents {
            let rawEvent = FSRawEvent(identifier: eventIds[index],
                                      path: paths[index],
                                      flags: eventFlags[index])
            
            if rawEvent.flags &
                UInt32(kFSEventStreamEventFlagItemCreated) |
                UInt32(kFSEventStreamEventFlagItemRenamed) |
                UInt32(kFSEventStreamEventFlagItemRemoved) != 0 {
                events.append(rawEvent)
            }
            
            let event = mySelf.process(rawEvents: events)
            callback(event)
        }
    }
    
    private func process(rawEvents: [FSRawEvent]) -> FSEvent {
        if rawEvents.count == 1, let rawEvent = rawEvents.first {
            return process(singleEvent: rawEvent)
        }
        
        // Для простоты предположим, что если пришло 2 события, значит это
        // переименование/перемещение, в реальном случае понадобятся более сложные проверки
        if rawEvents.count == 2 {
           return process(renameEvents: rawEvents)
        }
        
        return FSEvent()
    }
    
    private func process(singleEvent rawEvent: FSRawEvent) -> FSEvent {
        var event = FSEvent()
        event.sourcePath = rawEvent.path
        
        if rawEvent.flags & UInt32(kFSEventStreamEventFlagItemCreated) != 0 {
            event.type = .create
        }
        
        if rawEvent.flags & UInt32(kFSEventStreamEventFlagItemRemoved) != 0 {
            event.type = .remove
        }
        
        if rawEvent.flags & UInt32(kFSEventStreamEventFlagItemRenamed) != 0 {
            let fileExists = FileManager.default.fileExists(atPath: rawEvent.path)
            if !fileExists {
                event.type = .remove
            }
        }
        
        return event
    }
    
    private func process(renameEvents rawEvents: [FSRawEvent]) -> FSEvent {
        var event = FSEvent()
        let rawEvent1 = rawEvents[0]
        let rawEvent2 = rawEvents[1]
        let file1Exists = FileManager.default.fileExists(atPath: rawEvent1.path)
        let file2Exists = FileManager.default.fileExists(atPath: rawEvent2.path)
        
        event.type = .rename
        
        if file1Exists {
            event.destinationPath = rawEvent1.path
            event.sourcePath = rawEvent2.path
        } else if file2Exists {
            event.destinationPath = rawEvent2.path
            event.sourcePath = rawEvent1.path
        }
        
        return event
    }
    
    init(pathsToWatch: [String], callback: @escaping (FSEvent) -> Void) {
        self.pathsToWatch = pathsToWatch
        self.callback = callback
    }
    
    deinit {
        stop()
    }
    
    func start() {
        var context = FSEventStreamContext(version: 0,
                                           info: nil,
                                           retain: nil,
                                           release: nil,
                                           copyDescription: nil)
        context.info = Unmanaged.passUnretained(self).toOpaque()
        
        let flags = UInt32(kFSEventStreamCreateFlagUseCFTypes | kFSEventStreamCreateFlagFileEvents)
        
        stream = FSEventStreamCreate(kCFAllocatorDefault,
                                     eventCallback,
                                     &context,
                                     pathsToWatch as CFArray,
                                     FSEventStreamEventId(kFSEventStreamEventIdSinceNow),
                                     0,
                                     flags)
        
        FSEventStreamScheduleWithRunLoop(stream,
                                         CFRunLoopGetMain(),
                                         CFRunLoopMode.defaultMode.rawValue)
        FSEventStreamStart(stream)
    }
    
    func stop() {
        FSEventStreamStop(stream)
        FSEventStreamInvalidate(stream)
        FSEventStreamRelease(stream)
        stream = nil
    }
}
